/* This program aims at being a portable implementation of the iods */

#undef SHOWERR
#undef SHOWINFO

/**************************************************************************/
// Portable includes... I hope
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
/**************************************************************************/
// Less-portable includes...
#include <fcntl.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#if MSWIN
#include "win-compat.h"
#else /* MSWIN */
#include <sys/socket.h>
#include <netinet/in.h>
#endif /* !MSWIN */

#include "pread_pwrite.h"

/**************************************************************************/

#include "dbg.h"

// Macro used to display numerical IPs
// Reentrant as it does not use static buffers (like in inet_ntoa())
// Example: printf("IP of i=%d is " NTODSTR "\n", i, NTOD(x));
#define NTODSTR "%d.%d.%d.%d:%d"
#define NTOD(x) ((x).sin_addr.s_addr>>0)&0xFF, \
                ((x).sin_addr.s_addr>>8)&0xFF, \
                ((x).sin_addr.s_addr>>16)&0xFF, \
                ((x).sin_addr.s_addr>>24)&0xFF, \
                ntohs((x).sin_port)
/**************************************************************************/
// File: nfsp/default.h
#define MAXPENDINGFILES	10
#define NB_IODS_SUBDIR	0x20
#define BLOCKSIZE	(1<<16)
#define MAX_NEXTIODS  2

/**************************************************************************/
// File: nfsp/nfs_proto.h
#define NFS_MAXDATA 8192

enum nfsstat {
        NFS_OK = 0,
        NFSERR_PERM = 1,
        NFSERR_NOENT = 2,
        NFSERR_IO = 5,
        NFSERR_NXIO = 6,
        NFSERR_ACCES = 13,
        NFSERR_EXIST = 17,
        NFSERR_NODEV = 19,
        NFSERR_NOTDIR = 20,
        NFSERR_ISDIR = 21,
        NFSERR_INVAL = 22,
        NFSERR_FBIG = 27,
        NFSERR_NOSPC = 28,
        NFSERR_ROFS = 30,
        NFSERR_NAMETOOLONG = 63,
        NFSERR_NOTEMPTY = 66,
        NFSERR_DQUOT = 69,
        NFSERR_STALE = 70,
        NFSERR_WFLUSH = 99,
};
typedef enum nfsstat nfsstat;


enum ftype {
	NFNON = 0,
	NFREG = 1,
	NFDIR = 2,
	NFBLK = 3,
	NFCHR = 4,
	NFLNK = 5,
	NFSOCK = 6,
	NFBAD = 7,
	NFFIFO = 8,
};
typedef enum ftype ftype;


struct nfstime {
	u_int seconds;
	u_int useconds;
};
typedef struct nfstime nfstime;

struct fattr {
	ftype type;
	u_int mode;
	u_int nlink;
	u_int uid;
	u_int gid;
	u_int size;
	u_int blocksize;
	u_int rdev;
	u_int blocks;
	u_int fsid;
	u_int fileid;
	nfstime atime;
	nfstime mtime;
	nfstime ctime;
};
typedef struct fattr fattr;

struct readokres {
        fattr attributes;
        struct {
                u_int data_len;
                char *data_val;
        } data;
};
typedef struct readokres readokres;

struct readres {
        nfsstat status;
        union {
                readokres reply;
        } readres_u;
};
typedef struct readres readres;

struct attrstat {
        nfsstat status;
        union {
                fattr attributes;
        } attrstat_u;
};
typedef struct attrstat attrstat;

/**************************************************************************/
// File: nfsp/iodng.h
#define NFS_MAXDATA 8192
#define UDP_MAXLEN 65535


#define NFSP_MAGIC_MSG		0xBAFE
#define NFSP_READ_REQ		0x0001
#define NFSP_WRITE_REQ		0x0002
#define NFSP_WRITEMIRROR_REQ	0x0003
#define NFSP_FWDREAD_1_REQ	0x0010
#define NFSP_FWDREAD_2_REQ	0x0011
//#define NFSP_FWDWRITE_1_REQ	0x0020
//#define NFSP_FWDWRITE_2_REQ	0x0021
#define NFSP_PURGE_REQ		0x0030
#define NFSP_PING_REQ		0x0040
#define NFSP_PONG_REQ		0x0041
#define NFSP_MAX_REQ		0x0050


// The format of the messages used to communicate with iods.
// These messages must be delivered by means of UDP packets.
// WARNING: if the layout of this structure is changed
//          DO check the #define's below used to compute packet size
typedef struct {
	uint16_t magic;				// must be MAGIC_MSG_T
	uint16_t msgtype;			// one of the NFSP_*_REQ

//	char _MINIHEADER[0];

	uint32_t client_addr;			// client IP address
	uint32_t server_addr;			// server IP address
	uint16_t client_port;			// client port number
	uint16_t server_port;			// server port number

	unsigned char client_mac_addr[6];	// client MAC address

// char _STARTS[0];

	union {
		struct {
			u_long xid;			// same type as rpc_msg.rm_xid in rpc/rpc_msg.h
			u_int inode;			// the real inode of the metafile
			unsigned int seed;		// the seed of the file
			u_int stripe_bs;		// size in bytes of a storage block
			u_int stripe_w;			// stripe width (number of (v)iods used for storage)
			fattr attributes;		// file attributes as defined in nfs_prot.{x,h}

			u_int offset;			// offset of the requested operation

			u_int datalen;			// length of data to read or write
			//char _ENDS[0];
			char dataval[NFS_MAXDATA];	// nothing (size=0) if READ or data if WRITE
		} _RDWR;
		struct {
			u_long xid;			// same type as rpc_msg.rm_xid in rpc/rpc_msg.h

			u_int inode;			// the real inode of the metafile
			unsigned int seed;		// the seed of the file
			u_int stripe_bs;		// size in bytes of a storage block
			u_int stripe_w;			// stripe width (number of (v)iods used for storage)
			fattr attributes;		// file attributes as defined in nfs_prot.{x,h}
			u_int offset;			// offset of the requested operation

			// If idx == -1 then the iod must return the NFS ack to the client
			// If idx >=0 then the iod forwards the data to the
			//   next iod in the list that is nextiods[idx] (and
			//   decrease the idx in the message)
			//
			// Example:
			//   S: Serveur
			//   I1, I2, I3: iods
			//   In the message, nextiods[0]=I2 nextiods[1]=I3
			//
			//   S to I1:  idx=1   send to nextiods[idx] with idx set to 1-1
			//   I1 to I3: idx=0   send to nextiods[idx] with idx set to 0-1
			//   I3 to I2: idx=-1  send NFS ack to client
			int idx;					// still to mirror in 0..MAX_NEXTIODS-1
			struct sockaddr_in nextiods[MAX_NEXTIODS];	// nextiods[idx .. 0] are used

			u_int datalen;
		//	char _ENDS[0];
			char dataval[NFS_MAXDATA];	// nothing (size=0) if READ or data if WRITE
		} _WMR;

		struct {
			u_long xid;			// same type as rpc_msg.rm_xid in rpc/rpc_msg.h
			u_int inode;			// the real inode of the metafile
			unsigned int seed;		// the seed of the file
			u_int stripe_bs;		// size in bytes of a storage block
			u_int stripe_w;			// stripe width (number of (v)iods used for storage)
			fattr attributes;		// file attributes as defined in nfs_prot.{x,h}

			u_int offset;			// offset of the requested operation

			struct sockaddr_in nextiod;	// next hop (iod1 then null)
			u_int nextlen;			// next len to read (0 then what's still to read)

			u_int datalen;			// total length of data to read (NOT CHANGED)
		//	char _ENDS[0];
			char dataval[NFS_MAXDATA];	// void then DATA already read
		} _FWDRDWR;

		struct {
			int nfiles;
			struct {
				u_int inode;		// the real inode of the metafile
				unsigned int seed;	// the seed of file
			} files[MAXPENDINGFILES];
		//	char _ENDS[0];
		} _PURGE;

		struct {
			u_long seq;			// The sequence number
		//	char _ENDS[0];
		} _PING;

	} _KIND;
} msg_iod_t;

//#define rdwrreq _KIND._RDWR
#define readreq _KIND._RDWR
#define writereq _KIND._RDWR
#define writemirrorreq _KIND._WMR
//#define fwdrdwrreq _KIND._FWDRDWR
#define fwdreadreq _KIND._FWDRDWR
//#define fwdwritereq _KIND._FWDRDWR
#define purgereq _KIND._PURGE
#define pingreq _KIND._PING

#define OFFSETOF(TYPE, MEMBER)		((size_t) &((TYPE *)0)->MEMBER)
#define MSG_IOD_READ_LEN		(OFFSETOF(msg_iod_t,_KIND._RDWR.datavak))
#define MSG_IOD_WRITE_LEN(nbytes)	(OFFSETOF(msg_iod_t,_KIND._RDWR.dataval) + (nbytes))
#define MSG_IOD_WRITE_MIRROR_LEN(nbytes)	(OFFSETOF(msg_iod_t,_KIND._WMR.dataval) + (nbytes))
#define MSG_IOD_FWDREAD_LEN(nbytes)	(OFFSETOF(msg_iod_t,_KIND._FWDRDWR.dataval) + (nbytes))
//#define MSG_IOD_FWDWRITE_LEN(nbytes)	(OFFSETOF(msg_iod_t,_KIND._FWDRDWR.datavak) + (nbytes))
//#define MSG_IOD_PURGE_LEN		(OFFSETOF(msg_iod_t,_KIND._PURGE._ENDS))
#define MSG_IOD_PING_LEN		(OFFSETOF(msg_iod_t,_KIND._PING.seq) + sizeof(u_long))

//#define MSG_IOD_MINIHEADER_LEN		(OFFSETOF(msg_iod_t,client_addr))
//#define MSG_IOD_COMMON_LEN		(OFFSETOF(msg_iod_t,_STARTS))
#define MSG_IOD_LEN_MAX			(sizeof(msg_iod_t))



/**************************************************************************/
//File: /usr/include/rpc/rpc_msg.h
enum msg_type {
        CALL=0,
        REPLY=1,
};
enum reply_stat {
        MSG_ACCEPTED=0,
        MSG_DENIED=1,
};
enum accept_stat {
        SUCCESS=0,
        PROG_UNAVAIL=1,
        PROG_MISMATCH=2,
        PROC_UNAVAIL=3,
        GARBAGE_ARGS=4,
        SYSTEM_ERR=5
};
/**************************************************************************/
//File: /usr/include/rpc/types.h
typedef int enum_t;
/**************************************************************************/
static int udpsock = -1;
static struct sockaddr_in udpsock_in;
static int udpsock_port = 20000;

static int write_mirror_udpsock = -1;
static int forward_step1_read_udpsock = -1;
//static int tcpsocket = -1;
//static int tcpsock_port = 20000;


struct myrpc_msg {
	u_long rm_xid;
	enum msg_type rm_direction;
	enum reply_stat rp_stat;
	enum_t oa_flavor;					// must be NULL
	caddr_t oa_base;					// must be NULL
	enum accept_stat ar_stat;
};


// This function implements a one byte hash based on the inode and the seed and is used
// to find the 1st level directory in which is stored the file
static inline unsigned char hash(unsigned int inode, unsigned int seed) {
	unsigned int hash;
	hash = inode ^ seed;
	hash ^= hash >> 16;
	return (0xFF & hash) % NB_IODS_SUBDIR;
}



// path/file looks like XX/iXXXXXXXXsXXXXXXXX
#define KEYMAXSIZE (2+1+((1+8)*2)+1)

// This macro allows to define functions with extrachecks to compute the filename
//
#define DECLARE_key2fn(_internal_name,_msg_type,_msg_member)					\
static inline int key2fn_##_internal_name(const msg_iod_t * const msg, char * const filename) {	\
	 	int ret;									\
		if ((_msg_type) != msg->msgtype) {						\
			DISP("ERROR", "invalid request type=0x%04x instead of %x",			\
			msg->msgtype, (_msg_type));						\
			ret = -1;								\
		} else {									\
			(void) snprintf(filename, KEYMAXSIZE, "%2.2x/i%8.8xs%8.8x",		\
					hash(msg->_msg_member.inode, msg->_msg_member.seed),	\
					msg->_msg_member.inode, msg->_msg_member.seed);		\
			ret = 0;								\
		}										\
		return ret;									\
}

// Declare the key2fn_* functions
// For instance, the read function is:
//   static inline int key2fn_read(const msg_iod_t * const msg, char *filename);
DECLARE_key2fn(read,NFSP_READ_REQ,readreq)
DECLARE_key2fn(write,NFSP_WRITE_REQ,writereq)
DECLARE_key2fn(writemirror,NFSP_WRITEMIRROR_REQ,writemirrorreq)
DECLARE_key2fn(fwdread1,NFSP_FWDREAD_1_REQ,fwdreadreq)
DECLARE_key2fn(fwdread2,NFSP_FWDREAD_2_REQ,fwdreadreq)


static inline int get_realoffset(int bs, int w, int offset, int count) {
	int realoffset;

	DISP("INFO", "stripe_bs=%d stripe_w=%d o=%d count=%d", bs, w, offset, count);

	// the file will contain appended parts of stripes
	// node 0:    [ (0)   (N)   ...]
	// ...
	// node N-1:  [(N-1) (2N-1) ...]
	if (offset / bs != (offset + count - 1) / bs) {
		DISP("ERROR", "cross stripe operation requested");
		realoffset = -1;
	} else {
		realoffset = ((offset / (w * bs)) * bs) + (offset % bs);
	}

	return realoffset;
}



// To lower the load of the metaserver, data is sent the same way way it is read.
// Since on x86 we have the bad endianness (LSB) we have to convert the values
// to the good one(c)(tm) that is XDR's == MSB
//
// If data is already XDR compliant then this routine should be replaced by a
// simple: memmove(destination, source, sizeof(fattr))
//
//#warning "HTONL(FATTR) ENABLED"
static inline void copy_attributes(const fattr * destination, const fattr * const source) {
	unsigned int *src = (unsigned int*) source;
	unsigned int *dst = (unsigned int*) destination;
	unsigned int i = sizeof(fattr) / sizeof(unsigned int);	// number of 4byte words to copy
//	assert(sizeof(fattr) % sizeof(unsigned int) == 0);
	while (i--) {
		*dst = htonl(*src);
		src++;
		dst++;
	}
}


// Fills every field of 'rpc' using data provided by 'm'
static inline void do_rpc_header(struct myrpc_msg * const rpc, const u_long xid) {
	rpc->rm_xid = xid;
	rpc->rm_direction = htonl(REPLY);
	rpc->rp_stat = htonl(MSG_ACCEPTED);
	rpc->oa_flavor = 0;		// must be NULL
	rpc->oa_base = NULL;		// must be NULL
	rpc->ar_stat = htonl(SUCCESS);
}

static inline void do_send_udp(int s, const msg_iod_t *m, const char * const b, unsigned int size) {
	int ret;
	struct sockaddr_in sin;
	socklen_t sinlen = sizeof(sin);

	memset(&sin, '\0', sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = m->client_addr;
	sin.sin_port = m->client_port;
	ret = sendto(s, b, size, MSG_NOSIGNAL, (struct sockaddr *)&sin, sinlen);
	if (ret == -1) {
		DISP("ERROR", "sendto failed errno=%d", errno);
	}
}

static void do_reply_read(const msg_iod_t *m) {
	char fn[KEYMAXSIZE];
	char b[UDP_MAXLEN];
	struct myrpc_msg *rpc;
	struct readres *nfsres;
	char * endofdata;
	int fd;
	int tmp;


	if (-1 == key2fn_read(m, fn)) {
		return;
	}

	memset(b, '\0', UDP_MAXLEN);
	rpc = (struct myrpc_msg *) b;
	nfsres = (struct readres *) ((char*)rpc + sizeof(struct myrpc_msg));

	if (!m->readreq.datalen) {
		DISP("ERROR", "datalen is 0");
	}

	fd = open(fn, O_RDONLY);
	if (-1 == fd) {
		if (ENOENT == errno) {
			// We got a request but there is no file
			//  => this is a sparse file => we return an array of 0's
			DISP("INFO", "sparse file => return %d 0's for %s",
					m->readreq.datalen, fn);

			tmp = m->readreq.datalen;
			memset(&(nfsres->readres_u.reply.data.data_val), '\0', tmp);
			nfsres->status = NFS_OK;
			nfsres->readres_u.reply.data.data_len = htonl(tmp);
			if (tmp & 3) {
				// We have to pad the datas to a 4-bytes multiple
				int pad = 4 - (tmp & 3);
				memset(&(nfsres->readres_u.reply.data.data_val) + tmp, '\0', pad);
				tmp += pad;
			}

			copy_attributes(&(nfsres->readres_u.reply.attributes),
							&(m->readreq.attributes));
		} else {
			// XXX What to do ?
			DISP("ERROR", "cannot open read only errno=%d", errno);
			tmp = 0;
			nfsres->status = NFSERR_IO;
		}
	} else {
		int realoffset;
		int count = m->readreq.datalen;

		DISP("INFO", "reading %d bytes from %s", count, fn);

		realoffset = get_realoffset(m->readreq.stripe_bs, m->readreq.stripe_w, m->readreq.offset, count);

		if (realoffset != -1) {
			tmp = pread(fd, &(nfsres->readres_u.reply.data.data_val), count,
			    realoffset);
		} else {
			tmp = -1;
		}
		if (-1 == tmp) {
			// The read operation failed
			DISP("ERROR", "read failed");
			nfsres->status = NFSERR_IO;
			tmp = 0;

		} else {
			// read success
			nfsres->status = NFS_OK;
			nfsres->readres_u.reply.data.data_len = htonl(tmp);
			if (tmp & 3) {
				// We have to pad the datas to a 4-bytes multiple
				int pad = 4 - (tmp & 3);
				memset(&(nfsres->readres_u.reply.data.data_val) + tmp,
						'\0', pad);
				tmp += pad;
			}
			copy_attributes(&(nfsres->readres_u.reply.attributes),
							&(m->readreq.attributes));
		}
		if (-1 == close(fd)) {
			// Weird error...
			DISP("ERROR", "cannot close() errno=%d", errno);
		}
	}

	endofdata = (char*)(&(nfsres->readres_u.reply.data.data_val)) + tmp;

	// 2nd part : RPC header
	do_rpc_header(rpc, m->readreq.xid);

	do_send_udp(udpsock, m, b, endofdata - b);

}

static void do_forward_step1_read(msg_iod_t *m) {
        char fn[KEYMAXSIZE];            // will contain the filename
        int tmp;
        int ret;

        int readlen;
        struct sockaddr_in nextiod;
        int fd;

        if (-1 == key2fn_fwdread1(m, fn)) {
                return;
        }
        DISP("INFO", "*do_forward_step1_read()...  about to read from [%s]", fn);

        // Grab and sets values in m for forwarding
        readlen = m->fwdreadreq.datalen - m->fwdreadreq.nextlen;
        memmove(&nextiod, &(m->fwdreadreq.nextiod), sizeof(struct sockaddr_in));
        memset(&(m->fwdreadreq.nextiod), '\0', sizeof(struct sockaddr_in));

        fd = open(fn, O_RDONLY);
        if (-1 == fd) {
                if (ENOENT == errno) {
                        // We got a request but there is no file
                        //  => this is a sparse file => we return an array of 0's
                        DISP("INFO", "sparse file => return %d 0's for %s", readlen, fn);

                        memset(&(m->fwdreadreq.dataval[0]), '\0', readlen);
                        tmp = readlen;
                } else {
                        DISP("ERROR", "do_forward_step1_read(): cannot open read only errno=%d", errno);
                        tmp = 0;
                }
        } else {
                int realoffset;
                int count = readlen;
                DISP("INFO", "reading %d bytes from %s", count, fn);

                realoffset = get_realoffset(m->fwdreadreq.stripe_bs, m->fwdreadreq.stripe_w, m->fwdreadreq.offset, count);
                if (realoffset != -1) {
                        tmp = pread(fd, &(m->fwdreadreq.dataval[0]), count, realoffset);
                } else {
                        tmp = -1;
                }


                // read was not complete
                if (tmp != readlen) {
                        if (-1 == tmp) {
                                // read failure
                                DISP("ERROR", "do_forward_step1_read(): read failed errno=%d", errno);
                                tmp = 0;
                        } else {
                                // read success but EOF
                                DISP("ERROR", "do_forward_step1_read(): read only %d/%d bytes", tmp, readlen);
                                memset(&(m->fwdreadreq.dataval[tmp]), '\0', readlen - tmp);
                                tmp = readlen;
                        }
                }
                if (-1 == close(fd)) {
                        // Weird error...
                        DISP("ERROR", "do_forward_step1_read(): cannot close() errno=%d", errno);
                }
        }

        // Changing the type to a inter-iod forwarded read request
        m->msgtype = NFSP_FWDREAD_2_REQ;

        DISP("INFO", "do_forward_step1_read(): OK");
        ret = sendto(forward_step1_read_udpsock, (char *)m, MSG_IOD_FWDREAD_LEN(tmp), 0,
                        (struct sockaddr *)&nextiod, sizeof(struct sockaddr_in));
        if (ret == -1) {
                DISP("ERROR", "do_forward_step1_read(): cannot send UDP packet errno=%d", errno);
        }

}


static void do_forward_step2_read(msg_iod_t *m) {
	char fn[KEYMAXSIZE];
	char b[UDP_MAXLEN];
	struct myrpc_msg *rpc;
	struct readres *nfsres;
	char *readbuf;
	char *endofdata;

	int fd;
	int tmp;

	do {
		// XXX granted it's ugly...
		// we alter the packet received to get the name of the stripe
		u_int backup;
		backup = m->fwdreadreq.offset;
		m->fwdreadreq.offset += (m->fwdreadreq.datalen - m->fwdreadreq.nextlen);
		if (-1 == key2fn_fwdread2(m, fn)) {
			return;
		}
		m->fwdreadreq.offset = backup;
	} while(0);

	DISP("INFO", "*do_forward_step2_read()...  about to read from [%s]", fn);

#if 1|| defined(PARANOID)
	// Check that nextiod field has been Zero'ed
	do {
		struct sockaddr_in blank;
		memset(&blank, '\0', sizeof(struct sockaddr_in));
		if (memcmp(&(m->fwdreadreq.nextiod), &blank, sizeof(struct sockaddr_in))) {
			DISP("ERROR", "do_forward_step2_read: nextiod should be zero => ignoring req");
			return;
		}

		// If we're here then we should have a request spanning on several BLOCK
		if ((m->fwdreadreq.offset % BLOCKSIZE) + m->fwdreadreq.datalen <= BLOCKSIZE) {
			DISP("ERROR", "do_forward_step2_read: request does not span on several blocks => ignoring req");
			DISP("ERROR", "(offset[%d]%%BLOCKSIZE[%d]) + datalen[%d] = %d <= BLOCKSIZE",
				m->fwdreadreq.offset, BLOCKSIZE, m->fwdreadreq.datalen,
				(m->fwdreadreq.offset % BLOCKSIZE) + m->fwdreadreq.datalen);
			return;
		}

		// It's not critical but weird
		if (0 == m->fwdreadreq.datalen) {
			DISP("ERROR", "WARNING: datalen is 0");
		}

	} while(0);
#endif

	memset(b, '\0', UDP_MAXLEN);
	rpc = (struct myrpc_msg *) b;
	nfsres = (struct readres *) ((char*)rpc + sizeof(struct myrpc_msg));

	DISP("INFO", "Shifting by %d bytes", m->fwdreadreq.datalen - m->fwdreadreq.nextlen);
	// A payload of m->fwdreadreq.datalen - m->fwdreadreq.nextlen should be found
	// so the offset of the buffer is shifted by this value.
	memcpy(&(nfsres->readres_u.reply.data.data_val),
		   &(m->fwdreadreq.dataval),
		   m->fwdreadreq.datalen - m->fwdreadreq.nextlen);

	// readbuf is the buf in the message where we can complete the buffer
	readbuf = ((char *) &(nfsres->readres_u.reply.data.data_val))
		+ (m->fwdreadreq.datalen - m->fwdreadreq.nextlen);

	// 1st part : NFS Header
	// Note: at the end, tmp corresponds to the length of data to send back to the
	// client
	fd = open(fn, O_RDONLY);
	if (-1 == fd) {
		if (ENOENT == errno) {
			// Request but no file => sparse file => we return an array of 0's
			DISP("INFO", "sparse file => return %d 0's for %s", m->fwdreadreq.datalen, fn);

			tmp = m->fwdreadreq.datalen;
			memset(readbuf, '\0', m->fwdreadreq.nextlen);
			nfsres->status = NFS_OK;
			nfsres->readres_u.reply.data.data_len = htonl(tmp);
			if (tmp & 3) {
				// We have to pad the datas to a 4-bytes multiple
				int pad = 4 - (tmp & 3);
				memset(&(nfsres->readres_u.reply.data.data_val) + tmp, '\0', pad);
				tmp += pad;
			}

			copy_attributes(&(nfsres->readres_u.reply.attributes),
							&(m->fwdreadreq.attributes));

		} else {
			DISP("ERROR", "cannot open read only errno=%d", errno);
			nfsres->status = NFSERR_IO;
			tmp = 0;
		}
	} else {
		int realoffset;
		int count = m->fwdreadreq.nextlen;
		DISP("INFO", "reading count=nextlen=%d bytes from %s", count, fn);

		realoffset = get_realoffset(m->fwdreadreq.stripe_bs, m->fwdreadreq.stripe_w,
				m->fwdreadreq.offset + (m->fwdreadreq.datalen - m->fwdreadreq.nextlen),
				count);

		if (realoffset != -1) {
			tmp = pread(fd, readbuf, count, realoffset);
		} else {
			tmp = -1;
		}

		if (-1 == tmp) {
			// The read operation failed
			DISP("ERROR", "read failed");
			nfsres->status = NFSERR_IO;
			tmp = 0;
		} else {
			tmp += m->fwdreadreq.datalen - m->fwdreadreq.nextlen;
			// read success
			nfsres->status = NFS_OK;
			nfsres->readres_u.reply.data.data_len = htonl(tmp);
			if (tmp & 3) {
				// We have to pad the datas to a 4-bytes multiple
				int pad = 4 - (tmp & 3);
				memset(&(nfsres->readres_u.reply.data.data_val)+tmp, '\0', pad);
				tmp += pad;
			}
			copy_attributes(&(nfsres->readres_u.reply.attributes),
							&(m->readreq.attributes));
		}
		if (-1 == close(fd)) {
			DISP("ERROR", "cannot close() errno=%d", errno);
		}
	}

	DISP("INFO", "Full DATA size is %dB", tmp);
	endofdata = ((char *) &(nfsres->readres_u.reply.data.data_val)) + tmp;

	// 2nd part : RPC header
	do_rpc_header(rpc, m->readreq.xid);

	do_send_udp(udpsock, m, b, endofdata - b);
}


static void do_reply_writemirror(msg_iod_t *m) {
	char fn[KEYMAXSIZE];
	char b[UDP_MAXLEN];
	struct myrpc_msg *rpc;
	struct attrstat *nfsres;
	char *endofdata;
	int fd;
	int tmp;

	if (-1 == key2fn_writemirror(m, fn)) {
		return;
	}

	DISP("INFO", "*do_reply_writemirror()...  about to writemirror to [%s]", fn);

	// Sanity check: sensible index
	if (m->writemirrorreq.idx < -1 || m->writemirrorreq.idx > MAX_NEXTIODS - 1) {
		DISP("ERROR", "invalid request (idx <0 || idx > MAX_NEXTIODS-1)");
		return;
	}
	DISP("INFO", "*do_reply_writemirror()...  idx=%d", m->writemirrorreq.idx);

#if 0
	do {
		int i;

		for (i=0; i<MAX_NEXTIODS; i++) {
#warning "inet_aton is not MT-SAFE"
			DISP("INFO", "*do_reply_writemirror()... nextiods[%d] == %s:%d", i,
					inet_ntoa(m->writemirrorreq.nextiods[i].sin_addr),
					ntohs(m->writemirrorreq.nextiods[i].sin_port),
					MSG_IOD_PING_LEN);

		}
	} while (0);
#endif


	// Sanity check: a non-zero operation is wiser
	if (m->writemirrorreq.datalen == 0) {
		DISP("ERROR", "weird packet: datalen is 0 => request dropped");
		return;
	}

	// Sanity check: we do not handle writes when they're located on block boundaries
	if ((m->writemirrorreq.offset % BLOCKSIZE) + m->writemirrorreq.datalen
		> BLOCKSIZE) {
		DISP("ERROR", "(offset[%d]%%BLOCKSIZE[%d]) + datalen[%d] = %d > BLOCKSIZE",
			m->writemirrorreq.offset, BLOCKSIZE,
			m->writemirrorreq.datalen,
			(m->writemirrorreq.offset % BLOCKSIZE) + m->writemirrorreq.datalen);
		return;
	}



	memset(b, '\0', sizeof(b));
	rpc = (struct myrpc_msg *) b;
	nfsres = (struct attrstat *) ((char*)rpc + sizeof(struct myrpc_msg));

	// 1st part : NFS Header
	DISP("INFO", "Writing %d bytes to %s", m->writemirrorreq.datalen, fn);
	fd = open(fn, O_WRONLY|O_CREAT, 00600);
	if (-1 == fd) {
		DISP("ERROR", "cannot open %s WR|CREAT errno=%d", fn, errno);
		// XXX What to do ?
		tmp = -1;
	} else {
		int realoffset;
		int count = m->writemirrorreq.datalen;

		realoffset = get_realoffset(m->writemirrorreq.stripe_bs,
				m->writemirrorreq.stripe_w,
				m->writemirrorreq.offset, count);

		if (realoffset != -1) {
			tmp = pwrite(fd, m->writemirrorreq.dataval, count, realoffset);
		} else {
			tmp = -1;
		}
	}

	if (-1 == tmp) {
		// The write operation failed
		DISP("ERROR", "something failed during writemirror");
		nfsres->status = NFSERR_IO;
		// XXX what should be done ?
	} else {
		nfsres->status = NFS_OK;
		copy_attributes(&(nfsres->attrstat_u.attributes), &(m->writemirrorreq.attributes));
	}

	endofdata = (char*)&(nfsres->attrstat_u.attributes) + sizeof(fattr);

	if (-1 == close(fd)) {
		// Weird error... It might happen according to the man page but what to do in such
		// case ? Currently we can only do a warning as anything could have happened...
		DISP("ERROR", "cannot close() errno=%d", errno);
	}


	// We'll now decide what to do with this peculiar request
	// (Note: this comment is duplicated in iodng.h)
	//
	// If idx == -1 then the iod must return the NFS ack to the client
	// If idx >=0 then the iod forwards the data to the
	//   next iod in the list that is nextiods[idx] (and
	//   decrease the idx in the message)
	//
	// Example:
	//   S: Serveur
	//   I1, I2, I3: iods
	//   In the message, nextiods[0]=I2 nextiods[1]=I3
	//
	//   S to I1:  idx=1   send to nextiods[idx] with idx set to 1-1
	//   I1 to I3: idx=0   send to nextiods[idx] with idx set to 0-1
	//   I3 to I2: idx=-1  send NFS ack to client
	if (m->writemirrorreq.idx >= 0) {
		struct sockaddr_in sin;
		int sin_len = sizeof(sin);

		// there are still otherthe message has to be forwarded so we'll reuse the mirror
		// idx >= 1
		// NB: we use the source message to avoid a memory copy

		memcpy(&sin, &(m->writemirrorreq.nextiods[m->writemirrorreq.idx]), sizeof(sin));

		m->writemirrorreq.idx--;

		if (-1 == sendto(write_mirror_udpsock, (char *)m,
					MSG_IOD_WRITE_MIRROR_LEN(m->writemirrorreq.datalen),
					MSG_NOSIGNAL, (struct sockaddr *) &sin, sin_len)) {
			DISP("ERROR", "sendto failed on a UDP socket errno=%d", errno);
		}
		return;
	}

	if (nfsres->status != NFS_OK) {
		DISP("INFO", "a NFS error occured => stopping the forwarding and replying a NACK to the client");
	}
	if (m->writemirrorreq.idx == 0) {
		DISP("INFO", "final index so generating NFS reply");
	}

	// 2nd part : RPC header
	do_rpc_header(rpc, m->writemirrorreq.xid);

	do_send_udp(udpsock, m, b, endofdata - b);
}


static void do_reply_write(const msg_iod_t *m) {
	char fn[KEYMAXSIZE];
	char b[UDP_MAXLEN];
	struct myrpc_msg *rpc;
	struct attrstat *nfsres;
	char *endofdata;
	int fd;
	int tmp;

	if (-1 == key2fn_write(m, fn)) {
		return;
	}
	DISP("INFO", "*do_reply_write()...  about to write to [%s]", fn);

	// Zero-ing the headers
	memset(b, '\0', sizeof(b));

	// Buffer b will contain :
	//   IP Header / UDP Header / RPC Header / NFS Header
	rpc = (struct myrpc_msg *) b;
	nfsres = (struct attrstat *) ((char*)rpc + sizeof(struct myrpc_msg));

	// 1st part : NFS Header
	// Some sanity checks
	if (!m->writereq.datalen) {
		DISP("ERROR", "datalen is 0");
	}

	if ((m->writereq.offset % BLOCKSIZE) + m->writereq.datalen > BLOCKSIZE) {
		DISP("ERROR", "(offset[%d]%%BLOCKSIZE[%d]) + datalen[%d] = %d > BLOCKSIZE",
				m->writereq.offset,	BLOCKSIZE,
				m->writereq.datalen,
				(m->writereq.offset % BLOCKSIZE) + m->writereq.datalen);
	}

	DISP("INFO", "Writing %d bytes to %s", m->writereq.datalen, fn);
	fd = open(fn, O_WRONLY|O_CREAT, 00600);
	if (-1 == fd) {
		DISP("ERROR", "cannot open %s WR|CREAT errno=%d", fn, errno);
		// XXX What to do ?
		tmp = -1;
	} else {
		int realoffset;
		int count = m->writereq.datalen;

		realoffset = get_realoffset(m->writereq.stripe_bs,
				m->writereq.stripe_w, m->writereq.offset,
				count);

		if (realoffset != -1) {
			tmp = pwrite(fd, (char *) m->writereq.dataval, count, realoffset);
		} else {
			tmp = -1;
		}
	}

	if (-1 == tmp) {
		// The write operation failed
		DISP("ERROR", "write failed");
		nfsres->status = NFSERR_IO;
		// XXX what should be done ?
	} else {
		nfsres->status = NFS_OK;
		copy_attributes(&(nfsres->attrstat_u.attributes), &(m->writereq.attributes));
	}

	endofdata = (char*)&(nfsres->attrstat_u.attributes) + sizeof(fattr);

	if (-1 == close(fd)) {
		// Weird error...
		DISP("ERROR", "cannot close() errno=%d", errno);
	}

	// 2nd part : RPC header
	do_rpc_header(rpc, m->writereq.xid);

	do_send_udp(udpsock, m, b, endofdata - b);
}

static void do_purge(const msg_iod_t *m) {
	int idx;		// index of data file to remove in 'm'
	char file[KEYMAXSIZE];

	DISP("INFO", "*do_purge()...");

	for (idx = 0; idx < m->purgereq.nfiles; idx++) {
		unsigned int inode, seed;

		inode = m->purgereq.files[idx].inode;
		if (inode == 0) {
			// 0 is an invalid inode (and is used as padding)
			continue;
		}
		seed = m->purgereq.files[idx].seed;

		sprintf(file, "%2.2x/i%8.8xs%8.8x", hash(inode, seed), inode, seed);

		DISP("INFO", "*do_purge()...  about to unlink [%s]", file);

		if (-1 == unlink(file)) {
			DISP("ERROR", "error while unlinking '%s' errno=%d", file, errno);
		}
	}


}

static void do_reply_ping(const msg_iod_t *msrc) {
	char b[UDP_MAXLEN];
	msg_iod_t *msg;

	DISP("INFO", "*do_reply_ping_spoof()...");

	msg = (msg_iod_t *) b;

	// Setup of the PONG reply
	msg->magic = NFSP_MAGIC_MSG;
	msg->msgtype = NFSP_PONG_REQ;
	msg->client_addr = msrc->client_addr;
	msg->server_addr = msrc->server_addr;

	// Copy the MAC address
	memcpy(msg->client_mac_addr, msrc->client_mac_addr, 6);

	msg->client_port = msrc->client_port;
	msg->server_port = msrc->server_port;
	msg->pingreq.seq = msrc->pingreq.seq;


	do_send_udp(udpsock, msrc, b, MSG_IOD_PING_LEN);
}

static char * datapath = NULL;

void parse_command_line(int argc, char *argv[]) {

	DISP("INFO", "About to parse command line");


// XXX
	printf("\n");

	if (argc < 2) {
		printf("Error: not enough arguments\n");
		exit(1);
		}
	datapath = strdup(argv[1]);


}

//
int main(int argc, char *argv[])
{
	int i;

	#if MSWIN
	printf("Compiled under ms-windows");
	#else
	//DISP("INFO", "Compilation : " __DATE__ " at " __TIME__ " with " __VERSION__ "");
	DISP("INFO", "main(pid=%d): running as uid=%d euid=%d gid=%d egid=%d",
			getpid(), getuid(), geteuid(), getgid(), getegid());
	#endif

        // Check this function since it is parsed as we may still be r00t
        parse_command_line(argc, argv);

	// Chdir to the main dir
        if (-1 == chdir(datapath)) {
                DISP("ERROR", "cannot chdir() to %s", datapath);
		exit(1);
	}

        // Check if subdirs exist and are accessible
        for(i = 0; i < NB_IODS_SUBDIR; i++) {
                char curdir[KEYMAXSIZE];
                sprintf(curdir, "%2.2x", i);

#if MSWIN
		if (0 == mkdir(curdir)) {
                        DISP("ERROR", "Warning: created dir: %s", curdir);
                }
                strcat(curdir, "/.");
                if (-1 == access(curdir, 6)) {
                        DISP("ERROR", "Problem while accessing directory %s errno=%d", curdir, errno);
                        exit(1);
		}
#else /* MSWIN */
                if (0 == mkdir(curdir, 0700)) {
                        DISP("ERROR", "Warning: created dir: %s\n", curdir);
                }
                strcat(curdir, "/.");
		if (-1 == access(curdir, R_OK|W_OK|X_OK|F_OK)) {
                        DISP("ERROR", "Problem while accessing directory %s errno=%d", curdir, errno);
                        exit(1);
                }
#endif /* !MSWIN */
        }

	//----------------------------------------------------------
        // Setup the server socket on local port UDPPORT
        memset(&udpsock_in, '\0', sizeof(udpsock_in));
        udpsock_in.sin_family = AF_INET;
        udpsock_in.sin_addr.s_addr = INADDR_ANY;
        udpsock_in.sin_port = htons(udpsock_port);

        udpsock= socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
        if (udpsock == -1) {
                DISP("ERROR", "cannot open a UDP socket errno=%d", errno);
                exit(1);
        }
        if (-1 == bind(udpsock, (struct sockaddr *) &udpsock_in, sizeof(udpsock_in))) {
                DISP("ERROR", "cannot bind UDP socket %d errno=%d", udpsock, errno);
                exit(1);
        }

	while (1) {
		int recvd_bytes;
		struct sockaddr_in sinfrom;
		socklen_t sinfromlen = sizeof (sinfrom);
		msg_iod_t msg;
		msg_iod_t *msgp = &msg;

		DISP("INFO", "Waiting for UDP packet");

		recvd_bytes = recvfrom(udpsock, (char *)msgp, 65535, MSG_NOSIGNAL, (struct sockaddr *)&sinfrom, &sinfromlen);
		if (recvd_bytes == -1) {
			// recvfrom error
			DISP("ERROR", "errno = %d", errno);
			continue;
		}

		if (msgp->magic != NFSP_MAGIC_MSG) {
			// magic error
			DISP("ERROR", "not enough magic => dropped");
			continue;
		}

		switch (msgp->msgtype) {
			case NFSP_READ_REQ:
				do_reply_read(msgp);
				break;
			case NFSP_WRITE_REQ:
				do_reply_write(msgp);
				break;
			case NFSP_WRITEMIRROR_REQ:
				do_reply_writemirror(msgp);
				break;
			case NFSP_FWDREAD_1_REQ:
				// nfspd to iodng
				do_forward_step1_read(msgp);
				break;
			case NFSP_FWDREAD_2_REQ:
				// iodng to iodng
				do_forward_step2_read(msgp);
				break;

//			case NFSP_FWDWRITE_1_REQ:
//			case NFSP_FWDWRITE_2_REQ:
//				DISP("ERROR", "protocol error: received FWDWRITE_REQ - NYI");
//				break;

			case NFSP_PURGE_REQ:
				do_purge(msgp);
				break;

			case NFSP_PING_REQ:
				do_reply_ping(msgp);
				break;

			case NFSP_PONG_REQ:
				DISP("ERROR", "protocol error: received PONG_REP");
				break;

			default:
				DISP("ERROR", "unknown message type msgp->msgtype=%d",
						msgp->msgtype);
		}

	}

	return 0;
}
