// Author: Pierre Lombard <pierre.lombard@imag.fr>
// Date: 2001/2002
// License: GPLv2

#include <assert.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>

#include "timed_connect.h"

#define DEFFLAG 0
int
timed_setclientsock(const char * const host, const char * const service,
	      const char * const proto,unsigned long timeout_ms)
{
	struct hostent *phe;
	struct servent *pse;
	struct protoent *ppe;
	struct sockaddr_in server_in;
	int sock;
	int sock_type;

	memset(&server_in, '\0', sizeof (server_in));
	server_in.sin_family = AF_INET;

	if ((pse = getservbyname(service, proto)) != 0) {
		server_in.sin_port = htons((u_short) pse->s_port);
	} else if ((server_in.sin_port = htons((u_short) atoi(service))) == 0) {
		fprintf(stderr, "ERROR: service %s unavailable\n", service);
		exit(1);
	}

	if ((phe = gethostbyname(host)) != 0) {
		memcpy(&server_in.sin_addr, phe->h_addr, phe->h_length);
	} else if ((server_in.sin_addr.s_addr = inet_addr(host)) == INADDR_NONE) {
		fprintf(stderr, "ERROR: unknown host %s unavailable\n", host);
		exit(1);
	}

	if ((ppe = getprotobyname(proto)) == 0) {
		fprintf(stderr, "ERROR: protocol %s unavailable\n", proto);
		exit(1);
	}
	if (!strcmp(proto, "tcp")) {
		sock_type = SOCK_STREAM;
	} else if (!strcmp(proto, "udp")) {
		sock_type = SOCK_DGRAM;
	} else if (!strcmp(proto, "raw")) {
		sock_type = SOCK_RAW;
	} else {
		fprintf(stderr, "ERROR: protocol %s unavailable\n", proto);
		exit(1);
	}

	sock = socket(PF_INET, sock_type, ppe->p_proto);
	if (sock < 0) {
		fprintf(stderr, "ERROR: cannot create socket\n");
		exit(1);
	}

	switch (sock_type) {
	case SOCK_STREAM:
		if (0 > timed_connect(sock, (struct sockaddr *) &server_in, sizeof(server_in), timeout_ms)) {
			fprintf(stderr, "ERROR: cannot connect to %s on service/port %s\n", host, service);
			exit(1);
		}
		break;
		
	default:
		;
	}

	return sock;
}

int
setclientsock(const char * const host, const char * const service, const char * const proto) {
	return timed_setclientsock(host, service, proto, 0);
}

