// Author: Pierre Lombard <pierre.lombard@imag.fr>
// Date: 2001/2002
// License: GPLv2

#include <assert.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <netdb.h>

/* Starts a server (listening) socket */
int
setserversock(char *service, char *protocol, int queue_len)
{
	struct sockaddr_in server_in;
	struct servent *pse;
	struct protoent *ppe;

	int sock;
	int sock_type;

	memset((char *) &server_in, '\0', sizeof (server_in));
	server_in.sin_family = AF_INET;
	server_in.sin_addr.s_addr = INADDR_ANY;

	if ((pse = getservbyname(service, protocol)) != 0) {
		server_in.sin_port = htons((u_short) pse->s_port);
	} else if ((server_in.sin_port = htons((u_short) atoi(service))) == 0) {
		printf("ERR: service %s unavailable\n", protocol);
		exit(1);
	}

	if ((ppe = getprotobyname(protocol)) == 0) {
		printf("ERR: unknown protocol %s\n", protocol);
		exit(1);
	}

	if (!strcmp(protocol, "tcp")) {
		sock_type = SOCK_STREAM;
	} else if (!strcmp(protocol, "udp")) {
		sock_type = SOCK_DGRAM;
	} else if (!strcmp(protocol, "raw")) {
		sock_type = SOCK_RAW;
	} else {
		printf("ERR : unknown protocol %s\n", protocol);
		exit(1);
	}

	sock = socket(PF_INET, sock_type, ppe->p_proto);
	if (sock < 0) {
		printf("ERR : could not socket()\n");
		exit(1);
	}

	if (bind(sock, (struct sockaddr *) &server_in, sizeof (server_in)) < 0) {
		printf("ERR : could not bind()\n");
		exit(1);
	}

	if (sock_type == SOCK_STREAM) {
		if (listen(sock, queue_len) < 0) {
			printf("ERR : could not listen() on port %s\n",
			       service);
			exit(1);
		}
	}

	return sock;
}

static inline int
get_rcv_and_snd_buf(int sock, int type)
{
	int ret, opt, optsize;
	optsize = sizeof (opt);
	ret = getsockopt(sock, SOL_SOCKET, type, &opt, &optsize);
	assert(ret != -1);
	assert(optsize == sizeof (opt));
	return opt;
}

int
get_rcvbuf(int sock)
{
	return get_rcv_and_snd_buf(sock, SO_RCVBUF);
}

int
get_sndbuf(int sock)
{
	return get_rcv_and_snd_buf(sock, SO_SNDBUF);
}

static inline int
set_rcv_and_snd_buf(int sock, int size, int type)
{
	int ret;
	ret = setsockopt(sock, SOL_SOCKET, type, &size, sizeof (size));
	assert(ret != -1);
	return 0;
}

int
set_rcvbuf(int sock, int size)
{
	return set_rcv_and_snd_buf(sock, size, SO_RCVBUF);
}

int
set_sndbuf(int sock, int size)
{
	return set_rcv_and_snd_buf(sock, size, SO_SNDBUF);
}
